use crate::*;
use std::fmt;

#[derive(Debug, Clone)]
pub struct Reflector {
    pub variant: ReflectorVariant,
    mappings: Vec<Letter>,
}

impl Reflector {
    pub fn variant(variant: ReflectorVariant) -> Self {
        let mappings = match variant {
            ReflectorVariant::Identity => mappings(VARIANT_IDENTITY),
            ReflectorVariant::Beta => mappings(VARIANT_BETA),
            ReflectorVariant::Gamma => mappings(VARIANT_GAMMA),
            ReflectorVariant::A => mappings(VARIANT_A),
            ReflectorVariant::B => mappings(VARIANT_B),
            ReflectorVariant::C => mappings(VARIANT_C),
            ReflectorVariant::BThin => mappings(VARIANT_B_THIN),
            ReflectorVariant::CThin => mappings(VARIANT_C_THIN),
        };

        Reflector { variant, mappings }
    }

    pub fn all() -> Vec<Reflector> {
        vec![
            Reflector::variant(ReflectorVariant::A),
            Reflector::variant(ReflectorVariant::B),
            Reflector::variant(ReflectorVariant::C),
            Reflector::variant(ReflectorVariant::Beta),
            Reflector::variant(ReflectorVariant::Gamma),
            Reflector::variant(ReflectorVariant::BThin),
            Reflector::variant(ReflectorVariant::CThin),
            Reflector::variant(ReflectorVariant::Identity),
        ]
    }

    pub fn index(&self, letter: usize) -> usize {
        *self
            .mappings
            .get(letter)
            .expect(&format!("Reflector is missing mapping for a letter: {:?}", letter).to_string())
            as usize
    }

    pub fn letter(&self, letter: Letter) -> Letter {
        Letter::from(self.index(letter as usize))
    }
}

const VARIANT_IDENTITY: &'static str = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
const VARIANT_BETA: &'static str = "LEYJVCNIXWPBQMDRTAKZGFUHOS";
const VARIANT_GAMMA: &'static str = "FSOKANUERHMBTIYCWLQPZXVGJD";
const VARIANT_A: &'static str = "EJMZALYXVBWFCRQUONTSPIKHGD";
const VARIANT_B: &'static str = "YRUHQSLDPXNGOKMIEBFZCWVJAT";
const VARIANT_C: &'static str = "FVPJIAOYEDRZXWGCTKUQSBNMHL";
const VARIANT_B_THIN: &'static str = "ENKQAUYWJICOPBLMDXZVFTHRGS";
const VARIANT_C_THIN: &'static str = "RDOBJNTKVEHMLFCWZAXGYIPSUQ";

#[derive(Debug, Copy, Clone)]
pub enum ReflectorVariant {
    Identity,
    Beta,
    Gamma,
    A,
    B,
    C,
    BThin,
    CThin,
}

impl fmt::Display for ReflectorVariant {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            ReflectorVariant::Identity => f.write_str("Identity"),
            ReflectorVariant::Beta => f.write_str("Beta"),
            ReflectorVariant::Gamma => f.write_str("Gamma"),
            ReflectorVariant::A => f.write_str("A"),
            ReflectorVariant::B => f.write_str("B"),
            ReflectorVariant::C => f.write_str("C"),
            ReflectorVariant::BThin => f.write_str("B Thin"),
            ReflectorVariant::CThin => f.write_str("C Thin"),
        }
    }
}

fn mappings(input: &'static str) -> Vec<Letter> {
    let mut mappings = vec![];

    for c in input.chars() {
        mappings.push(Letter::from(c));
    }

    mappings
}
